@@ -1,5 +1,7 @@
 /*
- * Copyright (c) 2000-2005, 2007-2016
+ * SPDX-License-Identifier: ISC
+ *
+ * Copyright (c) 2000-2005, 2007-2019
  *	Todd C. Miller <Todd.Miller@sudo.ws>
  *
  * Permission to use, copy, modify, and distribute this software for any
@@ -19,6 +21,11 @@
  * Materiel Command, USAF, under agreement number F39502-99-1-0512.
  */
 
+/*
+ * This is an open source non-commercial project. Dear PVS-Studio, please check it.
+ * PVS-Studio Static Code Analyzer for C, C++ and C#: http://www.viva64.com
+ */
+
 #include <config.h>
 
 #include <sys/types.h>
@@ -54,42 +61,70 @@
  * Flags used in rebuild_env()
  */
 #undef DID_TERM
-#define DID_TERM	0x0001
+#define DID_TERM	0x00000001
 #undef DID_PATH
-#define DID_PATH	0x0002
+#define DID_PATH	0x00000002
 #undef DID_HOME
-#define DID_HOME	0x0004
+#define DID_HOME	0x00000004
 #undef DID_SHELL
-#define DID_SHELL	0x0008
+#define DID_SHELL	0x00000008
 #undef DID_LOGNAME
-#define DID_LOGNAME	0x0010
+#define DID_LOGNAME	0x00000010
 #undef DID_USER
-#define DID_USER    	0x0020
-#undef DID_USERNAME
-#define DID_USERNAME   	0x0040
+#define DID_USER    	0x00000020
+#undef DID_LOGIN
+#define DID_LOGIN   	0x00000040
 #undef DID_MAIL
-#define DID_MAIL   	0x0080
+#define DID_MAIL   	0x00000080
 #undef DID_MAX
-#define DID_MAX    	0x00ff
+#define DID_MAX    	0x0000ffff
 
 #undef KEPT_TERM
-#define KEPT_TERM	0x0100
+#define KEPT_TERM	0x00010000
 #undef KEPT_PATH
-#define KEPT_PATH	0x0200
+#define KEPT_PATH	0x00020000
 #undef KEPT_HOME
-#define KEPT_HOME	0x0400
+#define KEPT_HOME	0x00040000
 #undef KEPT_SHELL
-#define KEPT_SHELL	0x0800
+#define KEPT_SHELL	0x00080000
 #undef KEPT_LOGNAME
-#define KEPT_LOGNAME	0x1000
+#define KEPT_LOGNAME	0x00100000
 #undef KEPT_USER
-#define KEPT_USER    	0x2000
-#undef KEPT_USERNAME
-#define KEPT_USERNAME	0x4000
+#define KEPT_USER    	0x00200000
+#undef KEPT_LOGIN
+#define KEPT_LOGIN	0x00400000
 #undef KEPT_MAIL
-#define KEPT_MAIL	0x8000
+#define KEPT_MAIL	0x00800000
 #undef KEPT_MAX
-#define KEPT_MAX    	0xff00
+#define KEPT_MAX    	0xffff0000
+
+/*
+ * AIX sets the LOGIN environment variable too.
+ */
+#ifdef _AIX
+# define KEPT_USER_VARIABLES (KEPT_LOGIN|KEPT_LOGNAME|KEPT_USER)
+#else
+# define KEPT_USER_VARIABLES (KEPT_LOGNAME|KEPT_USER)
+#endif
+
+/*
+ * Functions to open, close and parse an environment file, either
+ * a system file such as /etc/environment or one specified in sudoers.
+ */
+struct sudoers_env_file {
+    void * (*open)(const char *);
+    void   (*close)(void *);
+    char * (*next)(void *, int *);
+};
+
+/*
+ * State for a local environment file.
+ */
+struct env_file_local {
+    FILE *fp;
+    char *line;
+    size_t linesize;
+};
 
 struct environment {
     char **envp;		/* pointer to the new environment */
@@ -569,11 +604,42 @@ static bool
 matches_env_list(const char *var, struct list_members *list, bool *full_match)
 {
     struct list_member *cur;
+    bool is_logname = false;
     debug_decl(matches_env_list, SUDOERS_DEBUG_ENV)
 
-    SLIST_FOREACH(cur, list, entries) {
-	if (matches_env_pattern(cur->value, var, full_match))
-	    debug_return_bool(true);
+    switch (*var) {
+    case 'L':
+	if (strncmp(var, "LOGNAME=", 8) == 0)
+	    is_logname = true;
+#ifdef _AIX
+	else if (strncmp(var, "LOGIN=", 6) == 0)
+	    is_logname = true;
+#endif
+	break;
+    case 'U':
+	if (strncmp(var, "USER=", 5) == 0)
+	    is_logname = true;
+	break;
+    }
+
+    if (is_logname) {
+	/*
+	 * We treat LOGIN, LOGNAME and USER specially.
+	 * If one is preserved/deleted we want to preserve/delete them all.
+	 */
+	SLIST_FOREACH(cur, list, entries) {
+	    if (matches_env_pattern(cur->value, "LOGNAME", full_match) ||
+#ifdef _AIX
+		matches_env_pattern(cur->value, "LOGIN", full_match) ||
+#endif
+		matches_env_pattern(cur->value, "USER", full_match))
+		debug_return_bool(true);
+	}
+    } else {
+	SLIST_FOREACH(cur, list, entries) {
+	    if (matches_env_pattern(cur->value, var, full_match))
+		debug_return_bool(true);
+	}
     }
     debug_return_bool(false);
 }
@@ -768,6 +834,10 @@ env_update_didvar(const char *ep, unsigned int *didvar)
 		SET(*didvar, DID_HOME);
 	    break;
 	case 'L':
+#ifdef _AIX
+	    if (strncmp(ep, "LOGIN=", 8) == 0)
+		SET(*didvar, DID_LOGIN);
+#endif
 	    if (strncmp(ep, "LOGNAME=", 8) == 0)
 		SET(*didvar, DID_LOGNAME);
 	    break;
@@ -790,8 +860,6 @@ env_update_didvar(const char *ep, unsigned int *didvar)
 	case 'U':
 	    if (strncmp(ep, "USER=", 5) == 0)
 		SET(*didvar, DID_USER);
-	    if (strncmp(ep, "USERNAME=", 9) == 0)
-		SET(*didvar, DID_USERNAME);
 	    break;
     }
 }
@@ -872,7 +940,8 @@ rebuild_env(void)
 #endif /* HAVE_LOGIN_CAP_H */
 #if defined(_AIX) || (defined(__linux__) && !defined(HAVE_PAM))
 	    /* Insert system-wide environment variables. */
-	    read_env_file(_PATH_ENVIRONMENT, true, false);
+	    if (!read_env_file(_PATH_ENVIRONMENT, true, false))
+		sudo_warn("%s", _PATH_ENVIRONMENT);
 #endif
 	    for (ep = env.envp; *ep; ep++)
 		env_update_didvar(*ep, &didvar);
@@ -900,7 +969,7 @@ rebuild_env(void)
 		env_update_didvar(*ep, &didvar);
 	    }
 	}
-	didvar |= didvar << 8;		/* convert DID_* to KEPT_* */
+	didvar |= didvar << 16;		/* convert DID_* to KEPT_* */
 
 	/*
 	 * Add in defaults.  In -i mode these come from the runas user,
@@ -910,21 +979,25 @@ rebuild_env(void)
 	if (ISSET(sudo_mode, MODE_LOGIN_SHELL)) {
 	    CHECK_SETENV2("SHELL", runas_pw->pw_shell,
 		ISSET(didvar, DID_SHELL), true);
+#ifdef _AIX
+	    CHECK_SETENV2("LOGIN", runas_pw->pw_name,
+		ISSET(didvar, DID_LOGIN), true);
+#endif
 	    CHECK_SETENV2("LOGNAME", runas_pw->pw_name,
 		ISSET(didvar, DID_LOGNAME), true);
 	    CHECK_SETENV2("USER", runas_pw->pw_name,
 		ISSET(didvar, DID_USER), true);
-	    CHECK_SETENV2("USERNAME", runas_pw->pw_name,
-		ISSET(didvar, DID_USERNAME), true);
 	} else {
 	    /* We will set LOGNAME later in the def_set_logname case. */
 	    if (!def_set_logname) {
+#ifdef _AIX
+		if (!ISSET(didvar, DID_LOGIN))
+		    CHECK_SETENV2("LOGIN", user_name, false, true);
+#endif
 		if (!ISSET(didvar, DID_LOGNAME))
 		    CHECK_SETENV2("LOGNAME", user_name, false, true);
 		if (!ISSET(didvar, DID_USER))
 		    CHECK_SETENV2("USER", user_name, false, true);
-		if (!ISSET(didvar, DID_USERNAME))
-		    CHECK_SETENV2("USERNAME", user_name, false, true);
 	    }
 	}
 
@@ -977,38 +1050,43 @@ rebuild_env(void)
     }
 
     /*
-     * Set $USER, $LOGNAME and $USERNAME to target if "set_logname" is not
+     * Set LOGIN, LOGNAME, and USER to target if "set_logname" is not
      * disabled.  We skip this if we are running a login shell (because
      * they have already been set).
      */
     if (def_set_logname && !ISSET(sudo_mode, MODE_LOGIN_SHELL)) {
-	if (!ISSET(didvar, (KEPT_LOGNAME|KEPT_USER|KEPT_USERNAME))) {
-	    /* Nothing preserved, set all three. */
+	if ((didvar & KEPT_USER_VARIABLES) == 0) {
+	    /* Nothing preserved, set them all. */
+#ifdef _AIX
+	    CHECK_SETENV2("LOGIN", runas_pw->pw_name, true, true);
+#endif
 	    CHECK_SETENV2("LOGNAME", runas_pw->pw_name, true, true);
 	    CHECK_SETENV2("USER", runas_pw->pw_name, true, true);
-	    CHECK_SETENV2("USERNAME", runas_pw->pw_name, true, true);
-	} else if ((didvar & (KEPT_LOGNAME|KEPT_USER|KEPT_USERNAME)) !=
-	    (KEPT_LOGNAME|KEPT_USER|KEPT_USERNAME)) {
+	} else if ((didvar & KEPT_USER_VARIABLES) != KEPT_USER_VARIABLES) {
 	    /*
-	     * Preserved some of LOGNAME, USER, USERNAME but not all.
+	     * Preserved some of LOGIN, LOGNAME, USER but not all.
 	     * Make the unset ones match so we don't end up with some
 	     * set to the invoking user and others set to the runas user.
 	     */
 	    if (ISSET(didvar, KEPT_LOGNAME))
 		cp = sudo_getenv("LOGNAME");
+#ifdef _AIX
+	    else if (ISSET(didvar, KEPT_LOGIN))
+		cp = sudo_getenv("LOGIN");
+#endif
 	    else if (ISSET(didvar, KEPT_USER))
 		cp = sudo_getenv("USER");
-	    else if (ISSET(didvar, KEPT_USERNAME))
-		cp = sudo_getenv("USERNAME");
 	    else
 		cp = NULL;
 	    if (cp != NULL) {
+#ifdef _AIX
+		if (!ISSET(didvar, KEPT_LOGIN))
+		    CHECK_SETENV2("LOGIN", cp, true, true);
+#endif
 		if (!ISSET(didvar, KEPT_LOGNAME))
 		    CHECK_SETENV2("LOGNAME", cp, true, true);
 		if (!ISSET(didvar, KEPT_USER))
 		    CHECK_SETENV2("USER", cp, true, true);
-		if (!ISSET(didvar, KEPT_USERNAME))
-		    CHECK_SETENV2("USERNAME", cp, true, true);
 	    }
 	}
     }
@@ -1044,9 +1122,9 @@ rebuild_env(void)
 
     /* Add the SUDO_USER, SUDO_UID, SUDO_GID environment variables. */
     CHECK_SETENV2("SUDO_USER", user_name, true, true);
-    snprintf(idbuf, sizeof(idbuf), "%u", (unsigned int) user_uid);
+    (void)snprintf(idbuf, sizeof(idbuf), "%u", (unsigned int) user_uid);
     CHECK_SETENV2("SUDO_UID", idbuf, true, true);
-    snprintf(idbuf, sizeof(idbuf), "%u", (unsigned int) user_gid);
+    (void)snprintf(idbuf, sizeof(idbuf), "%u", (unsigned int) user_gid);
     CHECK_SETENV2("SUDO_GID", idbuf, true, true);
 
     debug_return_bool(true);
@@ -1132,8 +1210,41 @@ validate_env_vars(char * const env_vars[])
     debug_return_bool(ret);
 }
 
+static void *
+env_file_open_local(const char *path)
+{
+    struct env_file_local *efl;
+    debug_decl(env_file_open_local, SUDOERS_DEBUG_ENV)
+
+    efl = calloc(1, sizeof(*efl));
+    if (efl != NULL) {
+	if ((efl->fp = fopen(path, "r")) == NULL) {
+	    if (errno != ENOENT) {
+		free(efl);
+		efl = NULL;
+	    }
+	}
+    }
+    debug_return_ptr(efl);
+}
+
+static void
+env_file_close_local(void *cookie)
+{
+    struct env_file_local *efl = cookie;
+    debug_decl(env_file_close_local, SUDOERS_DEBUG_ENV)
+
+    if (efl != NULL) {
+	if (efl->fp != NULL)
+	    fclose(efl->fp);
+	free(efl->line);
+	free(efl);
+    }
+    debug_return;
+}
+
 /*
- * Read in /etc/environment ala AIX and Linux.
+ * Parse /etc/environment lines ala AIX and Linux.
  * Lines may be in either of three formats:
  *  NAME=VALUE
  *  NAME="VALUE"
@@ -1142,24 +1253,27 @@ validate_env_vars(char * const env_vars[])
  * Invalid lines, blank lines, or lines consisting solely of a comment
  * character are skipped.
  */
-bool
-read_env_file(const char *path, bool overwrite, bool restricted)
+static char *
+env_file_next_local(void *cookie, int *errnum)
 {
-    FILE *fp;
-    bool ret = true;
-    char *cp, *var, *val, *line = NULL;
-    size_t var_len, val_len, linesize = 0;
-    debug_decl(read_env_file, SUDOERS_DEBUG_ENV)
+    struct env_file_local *efl = cookie;
+    char *var, *val, *ret = NULL;
+    size_t var_len, val_len;
+    debug_decl(env_file_next_local, SUDOERS_DEBUG_ENV)
 
-    if ((fp = fopen(path, "r")) == NULL) {
-	if (errno != ENOENT)
-	    ret = false;
-	debug_return_bool(ret);
-    }
+    *errnum = 0;
+    if (efl->fp == NULL)
+	debug_return_ptr(NULL);
+
+    for (;;) {
+	if (sudo_parseln(&efl->line, &efl->linesize, NULL, efl->fp, PARSELN_CONT_IGN) == -1) {
+	    if (!feof(efl->fp))
+		*errnum = errno;
+	    break;
+	}
 
-    while (sudo_parseln(&line, &linesize, NULL, fp, PARSELN_CONT_IGN) != -1) {
 	/* Skip blank or comment lines */
-	if (*(var = line) == '\0')
+	if (*(var = efl->line) == '\0')
 	    continue;
 
 	/* Skip optional "export " */
@@ -1178,15 +1292,6 @@ read_env_file(const char *path, bool overwrite, bool restricted)
 	var_len = (size_t)(val - var);
 	val_len = strlen(++val);
 
-	/*
-	 * If the env file is restricted, apply env_check and env_keep
-	 * when env_reset is set or env_delete when it is not.
-	 */
-	if (restricted) {
-	    if (def_env_reset ? !env_should_keep(var) : env_should_delete(var))
-		continue;
-	}
-
 	/* Strip leading and trailing single/double quotes */
 	if ((val[0] == '\'' || val[0] == '\"') && val[0] == val[val_len - 1]) {
 	    val[val_len - 1] = '\0';
@@ -1194,25 +1299,91 @@ read_env_file(const char *path, bool overwrite, bool restricted)
 	    val_len -= 2;
 	}
 
-	if ((cp = malloc(var_len + 1 + val_len + 1)) == NULL) {
+	if ((ret = malloc(var_len + 1 + val_len + 1)) == NULL) {
+	    *errnum = errno;
 	    sudo_debug_printf(SUDO_DEBUG_ERROR|SUDO_DEBUG_LINENO,
 		"unable to allocate memory");
-	    /* XXX - no undo on failure */
-	    ret = false;
+	} else {
+	    memcpy(ret, var, var_len + 1); /* includes '=' */
+	    memcpy(ret + var_len + 1, val, val_len + 1); /* includes NUL */
+	    sudoers_gc_add(GC_PTR, ret);
+	}
+	break;
+    }
+    debug_return_str(ret);
+}
+
+static struct sudoers_env_file env_file_sudoers = {
+    env_file_open_local,
+    env_file_close_local,
+    env_file_next_local
+};
+
+static struct sudoers_env_file env_file_system = {
+    env_file_open_local,
+    env_file_close_local,
+    env_file_next_local
+};
+
+void
+register_env_file(void * (*ef_open)(const char *), void (*ef_close)(void *),
+    char * (*ef_next)(void *, int *), bool system)
+{
+    struct sudoers_env_file *ef = system ? &env_file_system : &env_file_sudoers;
+
+    ef->open = ef_open;
+    ef->close = ef_close;
+    ef->next = ef_next;
+}
+
+bool
+read_env_file(const char *path, bool overwrite, bool restricted)
+{
+    struct sudoers_env_file *ef;
+    bool ret = true;
+    char *envstr;
+    void *cookie;
+    int errnum;
+    debug_decl(read_env_file, SUDOERS_DEBUG_ENV)
+
+    /*
+     * The environment file may be handled differently depending on
+     * whether it is specified in sudoers or the system.
+     */
+    if (path == def_env_file || path == def_restricted_env_file)
+	ef = &env_file_sudoers;
+    else
+	ef = &env_file_system;
+
+    cookie = ef->open(path);
+    if (cookie == NULL)
+	debug_return_bool(false);
+
+    for (;;) {
+	/* Keep reading until EOF or error. */
+	if ((envstr = ef->next(cookie, &errnum)) == NULL) {
+	    if (errnum != 0)
+		ret = false;
 	    break;
 	}
-	memcpy(cp, var, var_len + 1); /* includes '=' */
-	memcpy(cp + var_len + 1, val, val_len + 1); /* includes NUL */
 
-	sudoers_gc_add(GC_PTR, cp);
-	if (sudo_putenv(cp, true, overwrite) == -1) {
+	/*
+	 * If the env file is restricted, apply env_check and env_keep
+	 * when env_reset is set or env_delete when it is not.
+	 */
+	if (restricted) {
+	    if (def_env_reset ? !env_should_keep(envstr) : env_should_delete(envstr)) {
+		free(envstr);
+		continue;
+	    }
+	}
+	if (sudo_putenv(envstr, true, overwrite) == -1) {
 	    /* XXX - no undo on failure */
 	    ret = false;
 	    break;
 	}
     }
-    free(line);
-    fclose(fp);
+    ef->close(cookie);
 
     debug_return_bool(ret);
 }
