@@ -303,7 +303,7 @@ sudo_askpass(const char *askpass, const char *prompt)
     sa.sa_handler = SIG_DFL;
     (void) sigaction(SIGCHLD, &sa, &savechld);
 
-    if (pipe(pfd) == -1)
+    if (pipe2(pfd, O_CLOEXEC) == -1)
 	sudo_fatal(U_("unable to create pipe"));
 
     child = sudo_debug_fork();
@@ -311,13 +311,26 @@ sudo_askpass(const char *askpass, const char *prompt)
 	sudo_fatal(U_("unable to fork"));
 
     if (child == 0) {
-	/* child, point stdout to output side of the pipe and exec askpass */
-	if (dup2(pfd[1], STDOUT_FILENO) == -1) {
-	    sudo_warn("dup2");
-	    _exit(255);
+	/* child, set stdout to write side of the pipe or clear FD_CLOEXEC */
+	if (pfd[1] == STDOUT_FILENO) {
+	    if (fcntl(pfd[1], F_SETFD, 0) == -1) {
+		sudo_warn("fcntl");
+		_exit(255);
+	    }
+	} else {
+	    if (dup2(pfd[1], STDOUT_FILENO) == -1) {
+		sudo_warn("dup2");
+		_exit(255);
+	    }
 	}
 	if (setuid(ROOT_UID) == -1)
 	    sudo_warn("setuid(%d)", ROOT_UID);
+	/* Close fds before uid change to prevent prlimit sabotage on Linux. */
+	closefrom(STDERR_FILENO + 1);
+	/* Run the askpass program with the user's original resource limits. */
+	restore_limits();
+	/* But avoid a setuid() failure on Linux due to RLIMIT_NPROC. */
+	unlimit_nproc();
 	if (setgid(user_details.gid)) {
 	    sudo_warn(U_("unable to set gid to %u"), (unsigned int)user_details.gid);
 	    _exit(255);
@@ -326,7 +339,7 @@ sudo_askpass(const char *askpass, const char *prompt)
 	    sudo_warn(U_("unable to set uid to %u"), (unsigned int)user_details.uid);
 	    _exit(255);
 	}
-	closefrom(STDERR_FILENO + 1);
+	restore_nproc();
 	execl(askpass, askpass, prompt, (char *)NULL);
 	sudo_warn(U_("unable to run %s"), askpass);
 	_exit(255);
